# -*- perl -*-
#
#  Net::POP3Server - adpO - Extensible Perl POP3 server
#  
#  $Id: POP3Server.pm,v 1.4 2001/04/17 06:54:30 rhandom Exp $
#  
#  Copyright (C) 2001, Paul T Seamons
#                      paul@seamons.com
#                      http://seamons.com/
#  
#  This package may be distributed under the terms of either the
#  GNU General Public License 
#    or the
#  Perl Artistic License
#
#  All rights reserved.
#  
#  Please read the perldoc Net::POP3Server
#
################################################################

package Net::POP3Server;

use vars qw(@ISA $VERSION);
use strict;
use Net::Server::MultiType ();
use IO::Dir ();
use IO::File ();

$VERSION = '0.50';

@ISA = qw(Net::Server::MultiType);


### override-able options for this package
sub options {
  my $self = shift;
  my $prop = $self->{server};
  my $ref  = shift;

  $self->SUPER::options($ref);

  foreach ( qw(command_timeout session_timeout) ){
    $prop->{$_} = undef unless exists $prop->{$_};
    $ref->{$_} = \$prop->{$_};
  }

}


### check options
sub post_configure {
  my $self = shift;
  my $prop = $self->{server};

  ### let the parent do the rest
  $self->SUPER::post_configure;

  ### some default values to check for
  my $d = {command_timeout => 30,     # seconds to wait for command
           session_timeout => 5 * 60, # seconds to wait for entire session
         };
  foreach (keys %$d){
    $prop->{$_} = $d->{$_}
    unless defined($prop->{$_}) && $prop->{$_} =~ /^\d+$/;
  }
}


### what to do if IP is blocked
sub request_denied_hook {
  my $self = shift;

  sleep 5;
  print "-ERR Remote Address Denied.\r\n",
  $self->{time_start} = time;

  print "+OK - Goodbye.\r\n";
  close STDIN;

}


### main sub to read from STDIN
sub process_request {
  my $self = shift;
  my $prop = $self->{server};

  $self->{time_start} = time;
  $self->{user}       = undef;
  $self->{pass}       = undef;
  $self->{loggedIn}   = undef;
  $self->{got_messages} = undef;
  $self->{list}       = [];
  $self->{maildir}    = undef;
  $self->{is_open}    = 1;
  $self->{sequence}   = "";
  $self->{dele}       = {};
  $self->{lastmsg}    = 0;
  $self->{apop_key}   = $self->apop_key;

  ### be friendly
  print "+OK Welcome to $0. <$self->{apop_key}>\r\n";

  ### read from the user avoiding DoS
  local $SIG{ALRM} = sub { die "Timeout\n" };
  alarm $prop->{command_timeout};

  eval {
    while ( defined($_ = <STDIN>) ){

      ### make sure they don't keep going forever
      if( time - $self->{time_start} > $prop->{session_timeout} ){
        die "Timeout\n";
      }

      ### do the command
      $self->pop3_command( $_ );

      ### reset for another amount of time
      alarm $prop->{command_timeout};
    }
    alarm 0;
  };
  
  ### did we timeout ?
  if( $@ ){
    if( $@ =~ /timeout/i ){
      print "-ERR Timed out\r\n";
    }else{
      print "-ERR System Err\r\n";
    }
  }

}


### deal with a pop3 command
sub pop3_command {
  my $self = shift;

  ### trim the line
  my $line = shift;
  $line =~ s/\s*\n$//;
  $self->{sequence}.="[$line]";

  ### check and fix protocl
  $line =~ s/^\s*(\w{3,4})\b\s*//;
  my $command = $1;
  $command =~ tr/a-z/A-Z/;

  ### authorization phase
  if( $command eq 'STOP' ){
    $self->{server}->{max_requests} = 0;
    print "+OK - Stopping Server.\r\n";
    close STDIN;
  }elsif( $command eq 'QUIT' ){
    print "+OK - Thank you for coming.\r\n";
    close STDIN;
  }elsif( $command eq 'USER' ){
    return $self->USER( $line );
  }elsif( $command eq 'PASS' ){
    return $self->PASS( $line );
  }elsif( $command eq 'DBUG' ){
    return $self->DBUG( $line );
  }elsif( $command eq 'APOP' ){
    return $self->APOP( $line );
  }elsif( ! $self->{logged_in} ){
    return print "-ERR Login First.\r\n";
  }

  ### remaining command will need this call
  $self->get_messages;

  if( $command eq 'STAT' ){
    return $self->STAT( $line );
  }elsif( $command eq 'LIST' ){
    return $self->LIST( $line );
  }elsif( $command eq 'UIDL' ){
    return $self->UIDL( $line );
  }elsif( $command eq 'RETR' ){
    return $self->RETR( $line );
  }elsif( $command eq 'DELE' ){
    return $self->DELE( $line );
  }elsif( $command eq 'RSET' ){
    return $self->RSET( $line );
  }elsif( $command eq 'LAST' ){
    return print "+OK $self->{lastmsg}\r\n";;
  }elsif( $command eq 'NOOP' ){
    return print "+OK\r\n";
  }elsif( $command eq 'TOP' ){
    return $self->RETR( $line,'top' );
  }else{
    return print "-ERR Huh?[$command][$line]\r\n";
  }

}

sub DBUG {
  my $self=shift;
  require "Data/Dumper.pm";

  print "+OK Dumping object now...\r\n";
  print Data::Dumper::Dumper($self);
  print "\r\n.\r\n";
  return;
}

sub USER {
  my $self = shift;
  my $user = shift;

  if( $self->{logged_in} ){
    return print "-ERR Already logged in.\r\n";
  }elsif( defined($self->{user}) && length($self->{user}) ){
    return print "-ERR USER already supplied.\r\n";
  }

  $self->{user} = $user;
  return print "+OK Enter PASS.\r\n";
}

sub PASS {
  my $self = shift;
  my $pass = shift;

  if( $self->{logged_in} ){
    return print "-ERR Already logged in.\r\n";
  }
  if( !$self->{user} ){
    return print "-ERR First supply USER.\r\n";
  }

  ### check password
  if( ! $self->verify_password( $pass ) ){
    $self->{user} = undef;
    return print "-ERR Invalid USER or PASS.\r\n";
  }

  ### where is the mail directory
  $self->{maildir} = $self->get_maildir;
  if( ! -d $self->{maildir} ){
    return print "-ERR Maildir error\r\n";
  }

  ### all good
  $self->{logged_in} = 1;
  return print "+OK Logged In.\r\n";
}

### this should return whether the pass was correct or not
### default pass is foobar
sub verify_password {
  my $self = shift;
  my $user = $self->{user};
  my $pass = shift;

  return ($pass eq 'foobar');
}

### allow (or not) APOP commands
sub APOP {
  my $self = shift;
  my $line = shift;

  # return print "-ERR APOP Not supported.\r\n";

  ### check input
  if( $self->{logged_in} ){
    return print "-ERR Already logged in.\r\n";
  }
  if( $line !~ /^(\S+)\s+(\S+)$/ ){
    return print "-ERR Incorrect APOP format.\r\n";
  }
  my($user,$code) = ($1,$2);

  ### optain the apop secret
  my $pass = $self->apop_secret( $user );
  if( ! defined $pass ){
    return print "-ERR Invalid APOP login.\r\n";
  }

  ### do an md5 as per rfc
  require "Digest/MD5.pm";
  my $digest = Digest::MD5::md5_hex( "<$self->{apop_key}>$pass" );

  ### check the digest
  if( $code ne $digest ){
    return print "-ERR Invalid APOP login ($code)($digest).\r\n";
  }

  ### where is the mail directory
  $self->{user} = $user;
  $self->{maildir} = $self->get_maildir;
  if( ! -d $self->{maildir} ){
    return print "-ERR Maildir error\r\n";
  }

  ### all good
  $self->{logged_in} = 1;
  return print "+OK Logged In.\r\n";

}

### this should return the apop secret
### or undefined for invalid user
### default secret is foobar
sub apop_secret {
  my $self = shift;
  my $user = shift;

  return 'foobar';
}

### need a semi random key for apop
sub apop_key {
  my $self = shift;
  sprintf("%04d%.05f.%05d",
          time % 10000,
          rand(1),
          $$ % 10000 );
}


### default is to use a bogus directory
sub get_maildir {
  require "POSIX.pm";
  my $dir = POSIX::tmpnam();
  mkdir $dir;
  return $dir;
}


sub get_messages {
  my $self = shift;

  if( !$self->{got_messages} ){
    if( -d $self->{maildir} && opendir(_DIR,$self->{maildir}) ){
      my $file='';
      my $count=1;
      while( $file = readdir(_DIR) ){
        next if $file=~/^\.+$/;
        my $size = -s "$self->{maildir}/$file";
        $self->{list}->[$count++]="$size:$file";
      }
      closedir(_DIR);
    }
    $self->{got_messages}=1;
  }
}

sub STAT {
  my $self = shift;

  my $messages = 0;
  my $length   = 0;
  foreach ( @{$self->{list}} ){
    next unless /^(\d+):/; # skip 0 index and any deleted messages
    $length += $1;
    $messages++;
  }
  return print "+OK $messages $length\r\n";
}

sub LIST {
  my $self = shift;

  my $msg=shift;
  $msg=~s/\s//g;
  if( length($msg) ){
    if( $msg!~/^\d+$/ ){
      return print "-ERR Invalid Number.\r\n";
    }
    if( ! $self->{list}->[$msg] ){
      return print "-ERR Number out of range.\r\n";
    }
    $self->{list}->[$msg] =~ /^(\d+)/;
    return print "$msg $1\r\n";
  }

  print "+OK\r\n";
  foreach (my($x)=1; $x<=$#{$self->{list}}; $x++ ){
    $_=$self->{list}->[$x];
    next unless $_;
    /^(\d+)/;
    print "$x $1\r\n";
  }
  print ".\r\n";
  return;
}

sub get_uidl {
  my $self = shift;
  if( $_[0] =~ /^(\d+):(.+)$/ ){
    my ($size,$file)=($1,$2);
    my $time = ((-M "$self->{maildir}/$file")*86400+$^T)%(10**7);
    $file=~s|[A-Z]([\d\.]+)$|$1|; # (mailbox messages may contain a prefix letter)
    my $sum  = ascii_sum($file);
    ### just needs some unique identifier for this file (without opening it even)
    return "AG".length($file)."D".$time."N".int($size/3)."Q".$sum."X";
  }
  return "NOUNIQUEID [$_[0]]";
}

sub UIDL {
  my $self = shift;

  my $msg=shift;
  $msg=~s/\s//g;
  if( length($msg) ){
    if( $msg!~/^\d+$/ ){
      return print "-ERR Invalid Number.\r\n";
    }
    if( ! $self->{list}->[$msg] ){
      return print "-ERR Number out of range.\r\n";
    }
    my $uidl = $self->get_uidl( $self->{list}->[$msg] );
    return print "$msg $uidl\r\n";
  }

  print "+OK\r\n";
  foreach (my($x)=1; $x<=$#{$self->{list}}; $x++ ){
    $_=$self->{list}->[$x];
    next unless $_;
    my $uidl = $self->get_uidl( $self->{list}->[$x] );
    print "$x $uidl\r\n";
  }
  print ".\r\n";
  return;
}

sub RETR {
  my $self = shift;

  my $msg = shift;
  my $lines = 0;
  my $top = ($_[0] eq 'top');
  if( $top ){
    if( $msg !~ m/^\s*(\d+)\s+(\d+)/ ){
      return print "-ERR Protocol Error.\r\n";
    }
    ($msg,$lines)=($1,$2);
  }

  if( $msg!~/^\d+$/ ){
    return print "-ERR Invalid Number.\r\n";
  }
  if( ! $self->{list}->[$msg] ){
    return print "-ERR Number out of range.\r\n";
  }

  $self->{list}->[$msg] =~ /^\d+:(.+)$/;
  my $file="$self->{maildir}/$1";
  if( !-e $file ){
    return print "-ERR Internal Error - Missing File.\r\n";
  }
  if( !open(_RETR,"<$file") ){
    return print "-ERR Internal Error [$!].\r\n";
  }

  my $doing_body=0;
  $self->{lastmsg} = $msg if $msg > $self->{lastmsg};
  print "+OK\r\n";
  while( <_RETR> ){
    if( $top ){
      last if $doing_body && $lines--<=0;
      $doing_body=1 if /^[\r\n]+$/;
    }
    if( /^\./ ){
      $_=".".$_;
    }
    print $_;
  }
  close(_RETR);
  print  "\r\n.\r\n";

  return;
}

sub DELE {
  my $self = shift;

  my $msg = shift;
  if( $msg!~/^\d+$/ ){
    return print "-ERR Invalid Number.\r\n";
  }
  if( !$self->{list}->[$msg] ){
    return print "-ERR Number out of range.\r\n";
  }

  $self->{dele}->{$msg} = $self->{list}->[$msg];
  $self->{list}->[$msg] = '';

  return print "+OK Msg $msg marked\r\n";
}

sub RSET {
  my $self = shift;

  $self->{lastmsg} = 0;
  foreach my $msg ( keys %{$self->{dele}} ){
    $self->{list}->[$msg] = $self->{dele}->{$msg};
  }
  $self->{dele}={};

  return print "+OK Reset\r\n";
}


sub pre_server_close_hook {
  my $self=shift;
  return if $self->{logged}; # make sure I only log once, DESTROY will try also

  ### delete messages marked for deletion
  foreach my $msg ( keys %{$self->{dele}} ){
    $self->{dele}->{$msg} =~ /^\d+:(.+)$/;
    my $file="$self->{maildir}/$1";
    unlink $file if -e $file;
    if( -e $file ){
      $self->log(1,"Error: couldn't delete $file [$!]");
    }
  }

  my $end_time=time;
  my($sec,$min,$hour,$day,$mon)=localtime($self->{time_start});

  my $entry="$self->{server}->{peeraddr} - "
    . sprintf("[%02d%02d%02d%02d%02d]",$mon+1,$day,$hour,$min,$sec)
      . '['.($end_time-$self->{time_start}).']'
        . "$self->{sequence}\n";

  $self->log(4,$entry);

  $self->{logged}=1;

}

sub DESTROY {
  my $self=shift;
  $self->pre_server_close_hook unless $self->{logged};
}

### generic thing
sub ascii_sum {
  my $iter=length($_[0])-1;
  my $sum=0;
  $sum += ord(substr($_[0],$_,1)) for (0..$iter);
  return $sum;
}

1;

__END__

=head1 NAME

Net::POP3Server -- Post Office Protocol (3) server interface

=head1 B<SERVER SYNOPSIS>

  #!/usr/bin/perl
 
  package MyPOP3;

  use Net::POP3Server;

  @ISA = Net::POP3Server;

  MyPOP3->run();

  exit;

=cut

